import scallion.input._
import scallion.lexical._

object Lexer {

  private object definitions extends Lexers[Token, Char, Position] with CharRegExps {
    val lexer = Lexer(
      word("&&") |> { (cs, p) => Token.And(p._1) },
      word("||") |> { (cs, p) => Token.Or(p._1) },
      word("==>") |> { (cs, p) => Token.Implies(p._1) },
      word("def") |> { (cs, p) => Token.Def(p._1) },
      word("var") |> { (cs, p) => Token.Var(p._1) },

      word("if") |> { (cs, p) => Token.If(p._1) },
      word("else") |> { (cs, p) => Token.Else(p._1) },
      word("ensuring") |> { (cs, p) => Token.Ensuring(p._1) },
      word("invariant") |> { (cs, p) => Token.Invariant(p._1) },
      word("havoc") |> { (cs, p) => Token.Havoc(p._1) },
      word("assume") |> { (cs, p) => Token.Assume(p._1) },
      word("old") |> { (cs, p) => Token.Old(p._1) },

      elem('(') |> { (cs, p) => Token.Parenthesis(true, p._1) },
      elem(')') |> { (cs, p) => Token.Parenthesis(false, p._1) },

      elem('{') |> { (cs, p) => Token.CurlyBracket(true, p._1) },
      elem('}') |> { (cs, p) => Token.CurlyBracket(false, p._1) },

      elem('[') |> { (cs, p) => Token.SquareBracket(true, p._1) },
      elem(']') |> { (cs, p) => Token.SquareBracket(false, p._1) },

      word("BigInt") |> { (cs, p) => Token.BigIntType(p._1) },
      word("Boolean") |> { (cs, p) => Token.BoolType(p._1) },

      word(",") |> { (cs, p) => Token.Comma(p._1) },
      word(":") |> { (cs, p) => Token.Colon(p._1) },

      word("+") |> { (cs, p) => Token.Plus(p._1) },
      word("-") |> { (cs, p) => Token.Minus(p._1) },
      word("*") |> { (cs, p) => Token.Mult(p._1) },
      word("%") |> { (cs, p) => Token.Mod(p._1) },
      word("/") |> { (cs, p) => Token.Div(p._1) },

      word("!=") |> { (cs, p) => Token.NotEqual(p._1) },
      word("==") |> { (cs, p) => Token.DoubleEqual(p._1) },
      word("=") |> { (cs, p) => Token.Equal(p._1) },
      word("<=") |> { (cs, p) => Token.Leq(p._1) },
      word(">=") |> { (cs, p) => Token.Geq(p._1) },
      word("<") |> { (cs, p) => Token.Lt(p._1) },
      word(">") |> { (cs, p) => Token.Gt(p._1) },
      word("!") |> { (cs, p) => Token.Not(p._1) },

      word("true") |> { (cs, p) => Token.BooleanLiteral(true, p._1) },
      word("false") |> { (cs, p) => Token.BooleanLiteral(false, p._1) },

      many1(digit) |> { (cs, p) => Token.BigIntLiteral(BigInt(cs.mkString), p._1) },

      (elem(_.isLetter) | elem('_')) ~ many(elem(_.isLetterOrDigit) | elem('_')) |>
        { (cs, p) => Token.Identifier(cs.mkString, p._1) },

      many1(whiteSpace) |> { (cs, p) => Token.Ignore(p._1) }
    ).onError {
      case (cs, p) => Token.Error(cs.mkString, p._1)
    }
  }

  def apply(characters: Iterator[Char]): Iterator[Token] = {
    definitions.lexer(Source.fromIterator(characters, LinePositioner)).filter(Kind.of(_) != Kind.Ignore)
  }

  // def unapply(token: Token): String = token match {
  //   case Token.Not(_) => "!"
  //   case Token.And(_) => "&&"
  //   case Token.Or(_) => "||"
  //   case Token.Neg(_) => "-"
  //   case Token.Plus(_) => "+"
  //   case Token.Minus(_) => "-"
  //   case Token.Mult(_) => "*"
  //   case Token.Div(_) => "/"
  //   case Token.Mod(_) => "%"
  //   case Token.Parenthesis(true, _) => "("
  //   case Token.Parenthesis(false, _) => ")"
  //   case Token.Comma(_) => ","
  //   case Token.Equals(_) => "="
  //   case Token.BooleanLiteral(b, _) => b.toString
  //   case Token.Identifier(name, _) => name
  //   case Token.Error(content, _) => "<error: " + content + ">"
  //   case Token.Ignore(_) => "<ignored>"
  // }

  // def unapply(tokens: Seq[Token]): String = {

  //   def needsSpace(left: Kind, right: Kind): Boolean = (left, right) match {
  //     case (Kind.Identifier, Kind.Identifier) => true
  //     case (Kind.Dot, _) => true
  //     case (Kind.Comma, _) => true
  //     case (Kind.EquSign, _) => true
  //     case (_, Kind.EquSign) => true
  //     case (Kind.And, _) => true
  //     case (_, Kind.And) => true
  //     case (Kind.Or, _) => true
  //     case (_, Kind.Or) => true
  //     case (Kind.Implies, _) => true
  //     case (_, Kind.Implies) => true
  //     case (Kind.Iff, _) => true
  //     case (_, Kind.Iff) => true
  //     case _ => false
  //   }

  //   val spaces = tokens.zip(tokens.drop(1)).map {
  //     case (left, right) => if (needsSpace(Kind.of(left), Kind.of(right))) " " else ""
  //   }

  //   val strings = tokens.map(unapply(_))

  //   if (tokens.isEmpty) {
  //     ""
  //   }
  //   else {
  //     spaces.zip(strings.tail).foldLeft(strings.head) {
  //       case (acc, (space, string)) => acc + space + string
  //     }
  //   }
  // }
}