sealed trait Token {
  val position: Position
}

object Token {
  case class Not(position: Position) extends Token
  case class And(position: Position) extends Token
  case class Or(position: Position) extends Token
  case class Implies(position: Position) extends Token

  case class Neg(position: Position) extends Token
  case class Plus(position: Position) extends Token
  case class Minus(position: Position) extends Token
  case class Mult(position: Position) extends Token
  case class Div(position: Position) extends Token
  case class Mod(position: Position) extends Token

  case class Lt(position: Position) extends Token
  case class Leq(position: Position) extends Token
  case class Gt(position: Position) extends Token
  case class Geq(position: Position) extends Token

  case class If(position: Position) extends Token
  case class Else(position: Position) extends Token
  case class While(position: Position) extends Token
  case class Invariant(position: Position) extends Token
  case class Havoc(position: Position) extends Token
  case class Assume(position: Position) extends Token

  case class Parenthesis(isOpen: Boolean, position: Position) extends Token
  case class CurlyBracket(isOpen: Boolean, position: Position) extends Token
  case class SquareBracket(isOpen: Boolean, position: Position) extends Token

  case class BooleanLiteral(value: Boolean, position: Position) extends Token
  case class BigIntLiteral(value: BigInt, position: Position) extends Token

  case class Identifier(name: String, position: Position) extends Token

  case class Equal(position: Position) extends Token
  case class DoubleEqual(position: Position) extends Token
  case class NotEqual(position: Position) extends Token
  case class Def(position: Position) extends Token
  case class Old(position: Position) extends Token
  case class Var(position: Position) extends Token
  case class Ensuring(position: Position) extends Token

  case class BoolType(position: Position) extends Token
  case class BigIntType(position: Position) extends Token

  case class Colon(position: Position) extends Token
  case class Comma(position: Position) extends Token

  case class Ignore(position: Position) extends Token
  case class Error(error: String, position: Position) extends Token
}

sealed trait Kind

object Kind {
  case object Not extends Kind
  case object And extends Kind
  case object Or extends Kind
  case object Implies extends Kind

  case object Neg extends Kind
  case object Plus extends Kind
  case object Minus extends Kind
  case object Mult extends Kind
  case object Div extends Kind
  case object Mod extends Kind

  case object Lt extends Kind
  case object Leq extends Kind
  case object Gt extends Kind
  case object Geq extends Kind
  case object DoubleEqual extends Kind
  case object NotEqual extends Kind

  case object If extends Kind
  case object Else extends Kind
  case object While extends Kind
  case object Invariant extends Kind
  case object Havoc extends Kind
  case object Assume extends Kind

  case class Parenthesis(isOpen: Boolean) extends Kind
  case class CurlyBracket(isOpen: Boolean) extends Kind
  case class SquareBracket(isOpen: Boolean) extends Kind
  case object BooleanLiteral extends Kind
  case object BigIntLiteral extends Kind
  case object Identifier extends Kind

  case object Equal extends Kind
  case object Def extends Kind
  case object Var extends Kind
  case object Old extends Kind
  case object Ensuring extends Kind

  case object BoolType extends Kind
  case object BigIntType extends Kind

  case object Colon extends Kind
  case object Comma extends Kind
  case object Ignore extends Kind
  case object Error extends Kind

  def of(token: Token): Kind = token match {
    case Token.Not(_) => Kind.Not
    case Token.And(_) => Kind.And
    case Token.Or(_) => Kind.Or
    case Token.Implies(_) => Kind.Implies

    case Token.Neg(_) => Kind.Neg
    case Token.Plus(_) => Kind.Plus
    case Token.Minus(_) => Kind.Minus
    case Token.Mult(_) => Kind.Mult
    case Token.Div(_) => Kind.Div

    case Token.Lt(_) => Kind.Lt
    case Token.Leq(_) => Kind.Leq
    case Token.Gt(_) => Kind.Gt
    case Token.Geq(_) => Kind.Geq

    case Token.Mod(_) => Kind.Mod
    case Token.If(_) => Kind.If
    case Token.Else(_) => Kind.Else
    case Token.While(_) => Kind.While
    case Token.Invariant(_) => Kind.Invariant
    case Token.Havoc(_) => Kind.Havoc
    case Token.Assume(_) => Kind.Assume
    case Token.Old(_) => Kind.Old
    case Token.Ensuring(_) => Kind.Ensuring

    case Token.Parenthesis(isOpen, _) => Kind.Parenthesis(isOpen)
    case Token.CurlyBracket(isOpen, _) => Kind.CurlyBracket(isOpen)
    case Token.SquareBracket(isOpen, _) => Kind.SquareBracket(isOpen)
    case Token.BooleanLiteral(_, _) => Kind.BooleanLiteral
    case Token.BigIntLiteral(_, _) => Kind.BigIntLiteral
    case Token.Identifier(_, _) => Kind.Identifier
    case Token.Equal(_) => Kind.Equal
    case Token.DoubleEqual(_) => Kind.DoubleEqual
    case Token.NotEqual(_) => Kind.NotEqual
    case Token.Def(_) => Kind.Def
    case Token.Var(_) => Kind.Var

    case Token.BoolType(_) => Kind.BoolType
    case Token.BigIntType(_) => Kind.BigIntType

    case Token.Colon(_) => Kind.Colon
    case Token.Comma(_) => Kind.Comma
    case Token.Ignore(_) => Kind.Ignore
    case Token.Error(_, _) => Kind.Error
  }
}
