import java.io.{File, PrintWriter}
import java.nio.file.{Files, Path}
import scala.util.Using

import Expr._
import SMTInterface._
import Verification._

class Verification(p: Program) {
  def prime(id: Identifier) = Identifier(id.name + "'")

  println("Starting verification for program:")
  println(p)
  println("\n\n")

  val vds = p.variables
  val ids = vds.map(_.id)
  val toPrime = vds.map(vd => vd.id -> Var(prime(vd.id))).toMap

  var fresh = 1
  def freshen(id: Identifier) = Identifier(s"${id.name}#$fresh")

  def relation(s: Statement): Expr = {
    s match {
      case Assignment(x, e) =>
        and(
          Equals(Var(prime(x)), e) +:
          vds.filter(vd => vd.id != x).map(y => Equals(Var(prime(y.id)), Var(y.id)))
        )

      case Havoc(x) =>
        ???

      case NoOp =>
        ???

      case IfThenElse(e, c1, c2) =>
        ???

      case ProcedureInvocation(id, args) =>
        ???

      case Compose(c1, c2) =>
        val freshVars: Seq[Expr] = vds.map(vd => Var(freshen(vd.id)))
        fresh += 1
        ???

      case Assume(e) =>
        ???
    }

  }

  def generatePostVC(proc: ProcDef): Option[VC] = {
    proc.ensuring.map { post =>
      val rel = relation(proc.body)
      val formula = And(rel, Not(post.subst(toPrime)))
        .replace(vds.map(vd => Old(vd.id) -> Var(vd.id)).toMap)
      VC(proc.id, formula, "post-condition")
    }
  }

  def verifyVC(args: Seq[Arg], vc: VC): Boolean = {
    val dir = new File("smt-sessions")
    if (!dir.exists) Files.createDirectory(dir.toPath)

    val out = dir.toPath.resolve(s"${vc.name}.smt2").toFile

    val pw = new PrintWriter(out)

    exprToSMT(fresh, vds, args, vc.e, pw)

    invokeZ3(out)
  }

  def verify(): Boolean = {
    val vcs = p.procedures.values.flatMap(generatePostVC)

    val results = vcs.map { vc =>
      println(s"\nVerifying VC: ${vc.displayString}")
      println(vc.e)
      val res = verifyVC(p.procedures(vc.fid).args, vc)
      if (res) ok("VALID: " + vc.displayString)
      else error("INVALID: " + vc.displayString)
      res
    }
    results.foldLeft(true) { _ && _ }
  }
}

object Verification {
  def launch(config: Config): Unit = {
    verifyFile(config.file)
  }

  def verifyFile(f: File): Boolean = {
    (new Verification(Parser.parseFile(f))).verify()
  }

  def ok(msg: String): Unit    = println(Console.GREEN + msg + Console.RESET)
  def error(msg: String): Unit = println(Console.RED   + msg + Console.RESET)

  case class VC(fid: Identifier, e: Expr, descr: String) {
    def name: String = s"$fid-$descr"
    def displayString: String = s"VC $descr (in $fid)"
  }
}
